/*
====================================
Connections:
====================================
gpio 0 --> default uart transmit - dongle white wire
gpoi 1 --> default uart areceife - dongle green wire

gpio 10 spi1 sck --> FRAM SCK
gpio 11 spi1 sdo --> FRAM MOSI
gpio 12 spi1 sdi --> FRAM MISO
gpio 13 chip sel --> FRAM CS

FRAM is MB85RS4MT on ADAFRUIT 4719 breakout
https://cdn-shop.adafruit.com/product-files/4719/4719_MB85RS4MT.pdf

VBUS on PICO --> FRAM Vin  (5 volt)
FRAM WP connected to its 3v3 pin
FRAM HOLD connected to its 3v3 pin

The spi channel appears to run correctly at 20MHz on a white board.
So: 32 clocks to set up a read/write, then 8 clocks/byte read/write
This imples a max read/write rate of 2.5 MByte/sec for larger arrays
and 0.5 Mbyte/sec for single bytes.
(Probably runs at 30 MHz, but fails at 40 MHz. The FRAM is rated for
40 MHz, but the white board is not)
For integer writes max read/write rate is 0.6 MHz. Single int max is
0.3 MHz.

=====================================
FRAM low-level read/write support routines:
=====================================
0- void fram_spi_init()
---- sets up the spi channel

 1 - void fram_read_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t *buf, uint8_t len)
 ---- reads len bytes from array buf starting at addr
 
 2- void fram_write_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t * data, uint8_t len)
  ---- writes len bytes from array data starting at addr

3- void fram_write_enable(spi_inst_t *spi, uint cs_pin)
4- void fram_write_disable(spi_inst_t *spi, uint cs_pin)
---- enables/disables the write operation above

5- void fram_read_int(spi_inst_t *spi, uint cs_pin, uint32_t addr, int32_t *buf, uint8_t len)
---- reads len 32-bit integers from array buf starting at addr

6- void fram_write_int(spi_inst_t *spi, uint cs_pin, uint32_t addr, int32_t * data, uint8_t len)
---- writes len arbitrary type from array data starting at addr

7- void fram_read(spi_inst_t *spi, uint cs_pin, uint32_t addr, void * buf, uint8_t len)
---- reads len arbitrary type to array data from  FRAM starting at addr
---- BUT len has to reflect total number of bytes!

8- void fram_write(spi_inst_t *spi, uint cs_pin, uint32_t addr, void * data, uint8_t len)
---- writes len arbitrary type from array data to  FRAM starting at addr
---- BUT len has to reflect total number of bytes!

=====================================
FRAM higher level file routines
=====================================
From https://github.com/pholmes2012/Simple_FRAM_FileSystem

=========
volume routines
=========
void volume_create(int force)
The 'force' parameter should be set to true only if you are sure that you
want to format the device

int volume_free(void)
Returns the number of free bytes, or -1 if there is no volume defined.

int volume_file_count(void)
Returns number of defined files or -1 if there is no volume defined.

=========
file routines
=========

int fCreate(char* fileName, uint32_t maxSize)
Create a file with a name and a maximum size it can grow to.
The name can be any printable characters with max size 15 characters
Returns a file count or -1 if the file was not created.

int fOpen(char* fileName, open_file_index);         
Open an existing file, or return -1 if the file does not exist 
The file index should be in the range 0-9

 void fClose(file_open_index);                         
 Close an open file by saving new file size

 int fExists(char* fileName)
 Returns the volume file index if the file exists or -1 if it does not

 int fRead(int open_block_index, void * buf, uint8_t num_bytes)     
 Read in data from the current file position  whiich starts at 0 when the file is opened.
 updates open_block read pointer.
 File MUST be open

int fWrite(int open_block_index, void * data, uint8_t num_bytes)
Write out data starting at the current end of file.
Updates open_block file size.
File MUST be open

==========
to be implemented
==========

// file_open_index = fOpen(uint idx);               // Open a file at volumw index, or return false if fewer than idx+1 files exists

// fSeek(file_open_index, fileOffset);     // Seek to a position in a file, fail if out of bounds, return current position either way



=====================================
 threads on core0:
 1- the usual blinky and prints if there are lock errors
 2- FRAM test code
 
 */

#include "hardware/gpio.h"
#include "hardware/timer.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "hardware/spi.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include "hardware/sync.h"
#include "stdlib.h"

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// ==================================================
// === spi FRAM support 
// ==================================================
// file system specs from
// From https://github.com/pholmes2012/Simple_FRAM_FileSystem

//SPI configurations
#define PIN_SCK  10
#define PIN_SDO 11
#define PIN_SDIN 12
#define PIN_CS   13
#define SPI_PORT spi1

// chip select macros
#define CS_on  gpio_put(PIN_CS,0) 
#define CS_off gpio_put(PIN_CS,1) 

// FRAM commands
#define fram_write_cmd (2)
#define fram_read_cmd  (3)
#define fram_we_cmd    (6)
#define fram_wd_cmd    (4)

// === Initialize SPI channel ===
// (channel, baud rate set to 20MHz) 
void fram_spi_init(){
    spi_init(SPI_PORT, 20000000) ;

    // Map SPI signals to GPIO ports
    //gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SCK, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SDO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SDIN, GPIO_FUNC_SPI);
    // toggle CS manually rather than use framed CS
    gpio_init(PIN_CS) ;	
    // turn off chip sel
    gpio_put(PIN_CS, true) ;
    gpio_set_dir(PIN_CS, GPIO_OUT) ; 
}

// =====================
// === read commands ===
void fram_read_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t *buf, uint8_t len) {
    CS_on ;
    uint8_t cmdbuf[4] = {
            fram_read_cmd,
            addr >> 16,
            addr >> 8,
            addr
    };
    spi_write_blocking(spi, cmdbuf, 4);
    spi_read_blocking(spi, 0, buf, len);
    CS_off ; 
}

void fram_read_int(spi_inst_t *spi, uint cs_pin, uint32_t addr, int32_t *buf, uint8_t len) {
    CS_on ;
    uint8_t cmdbuf[4] = {
            fram_read_cmd,
            addr >> 16,
            addr >> 8,
            addr
    };
    //
    uint8_t intbuf[4*len] ;
    //
    spi_write_blocking(spi, cmdbuf, 4);
    spi_read_blocking(spi, 0, intbuf, 4*len);
    memcpy(buf, intbuf, 4*len) ;
    CS_off ; 
}

void fram_read(spi_inst_t *spi, uint cs_pin, uint32_t addr, void * buf, uint8_t len) {
    CS_on ;
    uint8_t cmdbuf[4] = {
            fram_read_cmd,
            addr >> 16,
            addr >> 8,
            addr
    };
    //
    spi_write_blocking(spi, cmdbuf, 4);
    spi_read_blocking(spi, 0, (uint8_t*) buf, len);
   //
    CS_off ; 
}

// === write control ===
void fram_write_enable(spi_inst_t *spi, uint cs_pin) {
    CS_on ; 
    uint8_t cmd = fram_we_cmd ;
    spi_write_blocking(spi, &cmd, 1);
    CS_off ; 
}

void fram_write_disable(spi_inst_t *spi, uint cs_pin) {
    CS_on ; 
    uint8_t cmd = fram_wd_cmd ;
    spi_write_blocking(spi, &cmd, 1);
    CS_off ; 
}

// === write commands ===
void fram_write_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t * data, uint8_t len) {
    uint8_t cmdbuf[4] = {
            fram_write_cmd,
            addr >> 16,
            addr >> 8,
            addr
    };
    fram_write_enable(SPI_PORT, cs_pin) ;
    CS_on ;
    spi_write_blocking(spi, cmdbuf, 4) ;
    spi_write_blocking(spi, data, len) ;
    CS_off ;
    fram_write_disable(SPI_PORT, cs_pin) ;
}

void fram_write_int(spi_inst_t *spi, uint cs_pin, uint32_t addr, int32_t * data, uint8_t len) {
    uint8_t cmdbuf[4] = {
            fram_write_cmd,
            addr >> 16,
            addr >> 8,
            addr
    };
    uint8_t intbuf[4*len] ;
    memcpy(intbuf, data, 4*len) ;
    //
    fram_write_enable(SPI_PORT, cs_pin) ;
    CS_on ;
    spi_write_blocking(spi, cmdbuf, 4) ;
    spi_write_blocking(spi, intbuf, 4*len) ;
    CS_off ;
    fram_write_disable(SPI_PORT, cs_pin) ;
}

void fram_write(spi_inst_t *spi, uint cs_pin, uint32_t addr, void * data, uint8_t len) {
    uint8_t cmdbuf[4] = {
            fram_write_cmd,
            addr >> 16,
            addr >> 8,
            addr
    };
    //
    fram_write_enable(SPI_PORT, cs_pin) ;
    CS_on ;
    spi_write_blocking(spi, cmdbuf, 4) ;
    spi_write_blocking(spi, (uint8_t*) data, len) ;
    CS_off ;
    fram_write_disable(SPI_PORT, cs_pin) ;
}

// ==========================================
// === FRAM file system definition
// ==========================================
// SPI attached FRAM memory

#define volume_size 524288
#define volume_max_file_count 100
#define volume_header_size sizeof(int)*3
#define volume_max_open_file_count 10
// an arbitrary integer for checking that the volume exists
#define volume_exists 4760
// start of the volume index
#define volume_file_index_base_location (volume_header_size+1)

// file directory item
#define file_name_max_size 16
// 3 ints plus the file name
#define volume_file_index_entry_size (sizeof(int)*3+file_name_max_size)
#define volume_file_index_size (volume_file_index_entry_size*volume_max_file_count)
// the beginning of the actual data for the first file
#define volume_zero_file_location (volume_file_index_size+volume_header_size+1)

// volume header:
// these items are written to  FRAM on volume creation
// these items are read from  FRAM on file creation
// vol_exists contains '4760' as a way to see if the volume already exists
struct vol_head {
    int vol_exists_number ;
    // init to volume_size - (vol header size + file directory size)
    int vol_free_space  ;
    // actually defined files
    int vol_file_count ;
} volume_head ;

// this struct is written to fram on file creation
// this struct is read during file opening
// this struct is writtein during file closeing
struct file_index {
    char file_name[file_name_max_size] ;
    int file_begin_location ;
    int file_max_size ;
    int file_current_size ;
} file_index_item ;

// data for one open file
struct open_file_item {
    int file_begin_location ;
    int file_current_size ; //(when opened)
    int file_max_size ;
    int  file_read_location ; //(init to 0 offset when opened)
    int file_write_location ; // (init to offset file_size+1)
    int index_location ; // position of file in volume index
    int valid ; // true while the file is actually open
} ;
// all of the open files
struct open_file_item open_files[volume_max_open_file_count] ;

// === Create the volume structure
// EVERY program should start by calling this with force=false
// force=true means always build it
// force=false means dont overwrite existing structure
void volume_create(int force){
    // read a volume header (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol if force is false
    if(volume_head.vol_exists_number != volume_exists || force == true){
        // build a new volume header
        volume_head.vol_exists_number = volume_exists ;
        volume_head.vol_free_space = volume_size - (volume_header_size + volume_file_index_size) ;
        volume_head.vol_file_count = -1 ;
        // load it to fram
        fram_write(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    }
    // error check
    else if(volume_head.vol_exists_number = volume_exists && force == false){
        printf("Volume create not done -- volume exists\n\r") ;
    }
}

// Return the size of free storage currently available for files
int volume_free(void){
    // read a volume header (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol 
    if(volume_head.vol_exists_number == volume_exists){
        return volume_head.vol_free_space  ;
    }
    else {
        return -1 ;
    }
}            

// volume_file_count();
// Return the size of free storage currently available for files
int volume_file_count(void){
    // read a volume header (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol if force is false
    if(volume_head.vol_exists_number == volume_exists){
        return volume_head.vol_file_count  ;
    }
    else {
        return -1 ;
    }
}    

// Create a file with a name and a maximum size it can grow to
// returns a file count
int fCreate(char* fileName, uint32_t maxSize) {
    // read the volume info and check for a volume
    // read a volume header (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol and check for avail space
    if((volume_head.vol_exists_number == volume_exists) && (volume_head.vol_free_space > maxSize)){
        // 
        volume_head.vol_file_count ++ ;
        // build new file descriptor
        strncpy(file_index_item.file_name, fileName, 16) ;
        file_index_item.file_begin_location = volume_size - volume_head.vol_free_space ; 
        file_index_item.file_max_size = maxSize ;
        file_index_item.file_current_size =-1 ;
        // write it
        int index_item_loc = volume_file_index_base_location + volume_head.vol_file_count * sizeof(file_index_item) ;
        fram_write(SPI_PORT, PIN_CS, index_item_loc, &file_index_item, sizeof(file_index_item)) ;

        // write new volume header
        // modify volume file count, free space
        volume_head.vol_free_space -= maxSize ;
        
        fram_write(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
        // testing
        // printf("fcreat: file count %d index loc %d name %s\n\r", volume_head.vol_file_count, index_item_loc, file_index_item.file_name) ;
        return volume_head.vol_file_count ;
    }
    else {  
        printf("fCreate: No volume, or not enough space\n\r") ;
        return -1 ;
    }   
}

// file_open_index = fOpen(char* fileName);         
// Open an existing file,  return negative if the file does not exist 
int fOpen(char* fileName, int unit_number){
    // get volume header for number of files
    // read a volume header (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol and check for avail space
    if((volume_head.vol_exists_number == volume_exists)){
        int n_files = volume_head.vol_file_count ;
        for(int i=0; i<=n_files; i++){
            int index_item_loc = volume_file_index_base_location + i * sizeof(file_index_item) ;
            fram_read(SPI_PORT, PIN_CS, index_item_loc, &file_index_item, sizeof(file_index_item)) ;
            // testing
            // printf("fopen: search# %d index loc %d name %s\n\r", i, index_item_loc, file_index_item.file_name) ;
            if(strcmp(fileName, file_index_item.file_name)==0){
                // new location
                //open_file_count++ ;
                // testing
                // printf("fopen file found: file name %s\n\r", file_index_item.file_name) ;
                // build open block
                open_files[unit_number].file_begin_location = file_index_item.file_begin_location ;
                open_files[unit_number].file_current_size = file_index_item.file_current_size; //(when opened)
                open_files[unit_number].file_max_size =  file_index_item.file_max_size;
                open_files[unit_number].file_read_location = 0 ; //(init to 0 offset when opened)
                open_files[unit_number].file_write_location = file_index_item.file_current_size += 1; // (init to offset file_size+1)
                open_files[unit_number].index_location = i ; // index of file index entry on fram
                open_files[unit_number].valid = true ; // index of file index entry on fram

                // return index to block
                return unit_number;
            }
        }
    }
    else {return -1;}
}

// fExists just checks for a given file name and returns the file index if true and -1 if false
int fExists(char* fileName){
    // get volume header for number of files
    // read a volume header (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol and check for avail space
    if((volume_head.vol_exists_number == volume_exists)){
        int n_files = volume_head.vol_file_count ;
        for(int i=0; i<=n_files; i++){
            int index_item_loc = volume_file_index_base_location + i * sizeof(file_index_item) ;
            fram_read(SPI_PORT, PIN_CS, index_item_loc, &file_index_item, sizeof(file_index_item)) ;
            // testing
            // printf("fopen: search# %d index loc %d name %s\n\r", i, index_item_loc, file_index_item.file_name) ;
            if(strcmp(fileName, file_index_item.file_name)==0){
                // return index to file block
                return i;
            }
        }
        return -1 ;
    }
    else {return -1;}
}

// fClose(open_block_index) -- copies open file block back to file index location on fram
// nulls the open_block
// any write operation is NOT finalized until the file is closed
void fClose(int open_block_index){
    // mark open block invalid
    open_files[open_block_index].valid = false ;
    // get current block from fram
    int index_item_loc = volume_file_index_base_location + open_files[open_block_index].index_location * sizeof(file_index_item) ;
    fram_read(SPI_PORT, PIN_CS, index_item_loc, &file_index_item, sizeof(file_index_item)) ;
    // modify length,
    file_index_item.file_current_size = open_files[open_block_index].file_current_size ;
    // testing
    //printf("fClose: size %d \n\r", open_files[open_block_index].file_current_size) ;
    // write it back
    fram_write(SPI_PORT, PIN_CS, index_item_loc, &file_index_item, sizeof(file_index_item)) ;
}

// fSeek(open_block_index, position, 'r' or 'w')
// file must be open so that it has a unit_number
// get the open_block contents from the unit_number
// modify read_pointer or write_pointer with check for sizes


// num_bytes = fRead(open_block_index, output_pointer, num_bytes)
// file must be open so that it has a unit_number
// get the open_block contents from the unit_number
// -- reads number of bytes at current read pointer with check for file size
// -- updates open_block read pointer
//
int fRead(int open_block_index, void * buf, uint32_t num_bytes){
    // is the file acetually open
    if(open_files[open_block_index].valid==false){
        printf("fread: File not open\n\r");
        return -1 ;
    }
    // need to check for file size
    // testing
    //printf("loccation and size %d %d\n\r", open_files[open_block_index].file_read_location, open_files[open_block_index].file_current_size);
    if (num_bytes+open_files[open_block_index].file_read_location <= open_files[open_block_index].file_current_size+1){
        // get absolute fram address
        int addr = open_files[open_block_index].file_begin_location + open_files[open_block_index].file_read_location ;
        uint8_t cmdbuf[4] = {
                fram_read_cmd,
                addr >> 16,
                addr >> 8,
                addr
        };
        //
        CS_on ;
        spi_write_blocking(SPI_PORT, cmdbuf, 4);
        spi_read_blocking(SPI_PORT, 0, (uint8_t*) buf, num_bytes);
        CS_off ; 
        // update the open file block -- write location -- file size
        open_files[open_block_index].file_read_location += num_bytes ;
        return num_bytes ;
    }
    else {
        printf("fRead: not enough data\n\r") ;
        return -1 ;
    }
}

// num_bytes = fWrite(open_block_index, input_pointer, num_bytes)
// file must be open so that it has a unit_number
// get the open_block contents from the unit_number
// -- writes at end of file pointer with check for max_size overflow
// -- updates open_block write pointer and file size
int fWrite(int open_block_index, void * data, uint32_t num_bytes){
    // is the file actually open
    if(open_files[open_block_index].valid==false){
        printf("fwrite: File not open\n\r");
        return -1 ;
    }
    // check for valid write size
    if (num_bytes+open_files[open_block_index].file_current_size <= open_files[open_block_index].file_max_size){
        // call the write
        // void fram_write(spi_inst_t *spi, uint cs_pin, uint32_t addr, void * data, uint8_t len)
        int addr = open_files[open_block_index].file_begin_location + open_files[open_block_index].file_write_location ;
        // tesing
        //printf("fWrite: addr %d\n\r", addr);
        //fram_write(SPI_PORT, PIN_CS, fram_address, (void *) input_pointer, num_bytes) ;
        uint8_t cmdbuf[4] = {
            fram_write_cmd,
            addr >> 16,
            addr >> 8,
            addr
        };
        //
        fram_write_enable(SPI_PORT, PIN_CS) ;
        CS_on ;
        spi_write_blocking(SPI_PORT, cmdbuf, 4) ;
        spi_write_blocking(SPI_PORT, (uint8_t*) data, num_bytes) ;
        CS_off ;
        fram_write_disable(SPI_PORT, PIN_CS) ;
        // update the open file block -- write location -- file size
        open_files[open_block_index].file_write_location += num_bytes ;
        open_files[open_block_index].file_current_size  += num_bytes ;
        return num_bytes ;
    }
    else {
        printf("fWrite: attempt to write past max size\n\r");
        return -1 ;
    }
}

// fReadAt(file_open_index,  fileOffset, *buffer, count); // Read in data after seeking to a file position
int fReadAt(int open_block_index, void * buf, uint32_t offset, uint32_t num_bytes){
    // is the file acetually open
    if(open_files[open_block_index].valid==false){
        printf("freadAt: File not open\n\r");
        return -1 ;
    }
    // need to check for file size
    // testing
    //printf("loccation and size %d %d\n\r", open_files[open_block_index].file_read_location, open_files[open_block_index].file_current_size);
    if (num_bytes+offset <= open_files[open_block_index].file_current_size+1){
        // get absolute fram address
        int addr = open_files[open_block_index].file_begin_location + offset ;
        uint8_t cmdbuf[4] = {
                fram_read_cmd,
                addr >> 16,
                addr >> 8,
                addr
        };
        //
        CS_on ;
        spi_write_blocking(SPI_PORT, cmdbuf, 4);
        spi_read_blocking(SPI_PORT, 0, (uint8_t*) buf, num_bytes);
        CS_off ; 
        // update the open file block -- write location -- file size
        // no location update for absolute read address
        //open_files[open_block_index].file_read_location += num_bytes ;
        return num_bytes ;
    }
    else {
        printf("fReadAt: not enough data\n\r") ;
        return -1 ;
    }
}

// NEED
// fWriteAt(file_open_index, fileOffset, *buffer, count); // Write out data after seeking to a file position 
int fWriteAt(int open_block_index, void * data, uint32_t location, uint32_t num_bytes){
    // is the file actually open
    if(open_files[open_block_index].valid==false){
        printf("fwriteAt: File not open\n\r");
        return -1 ;
    }
    // check for valid write size
    if (num_bytes+location <= open_files[open_block_index].file_current_size+1){
        // call the write
        // void fram_write(spi_inst_t *spi, uint cs_pin, uint32_t addr, void * data, uint8_t len)
        int addr = open_files[open_block_index].file_begin_location + location ;
        // tesing
        //printf("fWrite: addr %d\n\r", addr);
        //fram_write(SPI_PORT, PIN_CS, fram_address, (void *) input_pointer, num_bytes) ;
        uint8_t cmdbuf[4] = {
            fram_write_cmd,
            addr >> 16,
            addr >> 8,
            addr
        };
        //
        fram_write_enable(SPI_PORT, PIN_CS) ;
        CS_on ;
        spi_write_blocking(SPI_PORT, cmdbuf, 4) ;
        spi_write_blocking(SPI_PORT, (uint8_t*) data, num_bytes) ;
        CS_off ;
        fram_write_disable(SPI_PORT, PIN_CS) ;
        // update the open file block -- write location -- file size
        //open_files[open_block_index].file_write_location += num_bytes ;
        //if(num_bytes+location > open_files[open_block_index].file_current_size){
         //   open_files[open_block_index].file_current_size = num_bytes + location ;
        //}
        return num_bytes ;
    }
    else {
        printf("fWriteAt: attempt to write past current size\n\r");
        return -1 ;
    }
}

void volume_file_dir(void){
    // read a volume directory f (if any)
    fram_read(SPI_PORT, PIN_CS, 0, &volume_head, sizeof(volume_head)) ;
    // check for existing vol if force is false
    if(volume_head.vol_exists_number == volume_exists){
        printf("free bytes=%-6d     # files=%-2d \n\r", volume_free(), volume_file_count()+1);
        printf("#  name            size   max_size\n\r");
        int n_files = volume_head.vol_file_count ;
        for(int i=0; i<=n_files; i++){
            int index_item_loc = volume_file_index_base_location + i * sizeof(file_index_item) ;
            fram_read(SPI_PORT, PIN_CS, index_item_loc, &file_index_item, sizeof(file_index_item)) ;
            // testing
            // printf("fopen: search# %d index loc %d name %s\n\r", i, index_item_loc, file_index_item.file_name) ;
            printf("%-2d %-15s %-6d %-6d\n\r", i, file_index_item.file_name, file_index_item.file_current_size, file_index_item.file_max_size) ;
        } 
    }
    else {
        printf("No directory\n\r") ;
    }
}    

// Needs:
// Osize(file_open_index);             // Return the current size of the file
// OreadLocation(file_open_index);
// OWriteLocation(file_open_index);
// OSizeMax(file_open_index);          // Return the maximum size the file can be


// ==================================================
// === user's serial input thread
//  testing FRAM file commands
// ==================================================
// serial_read and serial_write do not block any thread
// except this one

/* file command list
format  --  build a new volume
create  --  make an empty file
open    --  open an existing file -- file MUST be open for read/write
close   --  close an open file -- no file write is permanent until closed
read    --  read an open file from the currnet location
readat  --  read an open file at the specifed file byte
write   --  write at the end of an open file
writeat --  write at the specifed file byte
dir     --  print volume and file info

For this example, only single integers are written for testing
and to understand the logical dependencies of the file system.
NOTE that all reported and required length parameters are in BYTES!
Therefore, since we are writing integers a file containing 3 integers
will report a size of 11 (zero-based). To use readAt you would need to
specify an offset of 0, 4, or 8 for int items 0, 1, 2.
You might typically
1. format the volume
2. create a file
3. open it
4. write to it
5. write again
6. close
7. -- reset the processor -- 
8. use dir to see persistent file is still there
9. open it again
10. read it

*/

static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      static int num_in1, num_in2, num_in3 ;
      static char file_cmd[32], c_buf, file_name[16];
      //
      // start up SPI channel
      fram_spi_init();
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "file cmd: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        // convert input string to number
        sscanf(pt_serial_in_buffer,"%s", &file_cmd) ;

        // parse the command
        if(strcmp(file_cmd,"format")==0){
            sprintf(pt_serial_out_buffer, "Are you sure? y/n: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%c", &c_buf) ;
            // if yes, force a volume create
            if(c_buf == 'y') volume_create(true) ;
        }

        else if(strcmp(file_cmd,"create")==0) {
              sprintf(pt_serial_out_buffer, "file_name, max_size: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%s %d", file_name, &num_in1) ;
            // if yes, force a volume create
            if(fExists(file_name) < 0) {
                fCreate(file_name, num_in1);
            } else {
                printf("File aready exists\n\r");
            }
        }

        else if(strcmp(file_cmd,"open")==0) {
              sprintf(pt_serial_out_buffer, "file_name, desired_open_file_id: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%s %d", file_name, &num_in1) ;
            // if yes, force a volume create
            if(fExists(file_name) > -1) {
                fOpen(file_name, num_in1);
            } else {
                printf("File does not exist\n\r");
            }
        }

        else if(strcmp(file_cmd,"close")==0) {
              sprintf(pt_serial_out_buffer, "open_file_id: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%d", &num_in1) ;
            // if yes, force a volume create
            if(open_files[num_in1].valid == true) {
                fClose(num_in1);
            } else {
                printf("File not open\n\r");
            }
        }

        else if(strcmp(file_cmd,"write")==0) {// write FRAM, command form: 
            //  open_file_number integer_data
            //int fWrite(open_file_0, fram_data, 4*sizeof(int)) ;
            sprintf(pt_serial_out_buffer, "open_file_identifer, int_data: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%d %d", &num_in1, &num_in2) ;
            if(open_files[num_in1].valid == false) {
                printf("File not open\n\r");
            }
            else {
                fWrite(num_in1, &num_in2, sizeof(int)) ;
            }
        }

        else if(strcmp(file_cmd,"read")==0) {
            //  open_file_number integer_data_r
            sprintf(pt_serial_out_buffer, "open_file_identifer: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%d", &num_in1) ;
            // int fRead(int open_block_index, void * buf, uint8_t num_bytes)
            if(open_files[num_in1].valid == false) {
                printf("File not open\n\r");
            }
            else {
                fRead(num_in1, &num_in2, sizeof(int)) ;
                printf("FRAM data = %d\n\r", num_in2);
            }
        }

        else if(strcmp(file_cmd,"readAt")==0) {
            //  open_file_number integer_data_r
            sprintf(pt_serial_out_buffer, "open_file_identifer, location: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%d %d", &num_in1, &num_in2) ;
            // int fRead(int open_block_index, void * buf, uint8_t num_bytes)
            if(open_files[num_in1].valid == false) {
                printf("File not open\n\r");
            }
            else {
                fReadAt(num_in1, &num_in3, num_in2, sizeof(int)) ;
                printf("FRAM data = %d\n\r", num_in3);
            }
        }

        else if(strcmp(file_cmd,"writeAt")==0) {// write FRAM, command form: 
            //  open_file_number integer_data
            //int fWrite(open_file_0, fram_data, location, 4*sizeof(int)) ;
            sprintf(pt_serial_out_buffer, "open_file_identifer, location, int_data: ");
            // spawn a thread to do the non-blocking write
            serial_write ;
            // spawn a thread to do the non-blocking serial read
            serial_read ;
            // convert input string to number
            sscanf(pt_serial_in_buffer,"%d %d %d", &num_in1, &num_in2, &num_in3) ;
            if(open_files[num_in1].valid == false) {
                printf("File not open\n\r");
            }
            else {
                fWriteAt(num_in1, &num_in3, num_in2, sizeof(int)) ;
            }
        }

         else if(strcmp(file_cmd,"dir")==0) {
            volume_file_dir();
         }

        else {
            printf("???\n\r");
            printf("Cmds: format, create, open, close, write, read, readAt, writeAt, dir \n\r") ;
        }


        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  
  //  === add threads  ====================
  // for core 1
  //pt_add_thread(protothread_toggle_fast);
  //pt_add_thread(protothread_toggle3);
  //pt_add_thread(protothread_sum) ;
  //
  // === initalize the scheduler ==========
  //pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.1 two-core\n\r");
  
  // start core 1 threads
  //multicore_reset_core1();
  //multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //pt_add_thread(protothread_fram);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////